package cw.GI;

import cw.log.Log;


/**
 * This class simply performs a call to the gamemanager
 * at regular intervals, monitoring execution time in order
 * to run the game as smoothly as possible.
 */
public class GIThread implements Runnable 
{

  public int mClockCnt; // holds the main clock cnt 
  int mClockDelay; // holds the clock speed
  
  boolean noDelay=false; // used to test program speed wo sleeping
  
  boolean mBusyFlag; // flag to indicate that thread is going through step function
  GIHarness mHarness; // reference to the harness
    
  public Object mSemaphore; 
  // semaphore to provide exclusion between step() called from this thread 
  // and paint() called from outside 
  
  /**
   * constructor
   */
  GIThread (GIHarness pHarness)
  {
   //Nog.msg('I',"GIThread constructor()");
    mClockCnt=0;
    mClockDelay=200;
    mHarness=pHarness;
    mSemaphore=new Object();
  }
 
 
  // stats for monitoring program speed
 // public long mStatCanceledAddWord=0; // number of times the additional word (typically paint) has been canceled because this thread was not sleeping
  public long mStatSleptTime=0; // total slept time
  
  public long mStatOverloadTime=0; // total overload time
  public long mStatOverloadTimes=0;  // number of times an overload occured
  public long mStatTotalTime=0; // the actual total time
  public long mStatClockCnt=0;  // number of cycles 
  public long mStatWorkTime=0; // total work time
  //public long mStatAddWorkTime=0;
  
  public void initStat()
  {
    mStatTotalTime=0;
    mStatSleptTime=0;
    mStatOverloadTime=0;
    mStatOverloadTimes=0;
    mStatClockCnt=0;
    mStatWorkTime=0;
    //mStatAddWorkTime=0;
   // mStatCanceledAddWord=0;
    
  
  } 
  
 // public void addStatWorkTime(long millis)
  //{
  ////}
  
  
  public String getStatString()
  {
  //Nog.msg('D',"getStatString():");
    
   //Nog.msg('D',"   mStatTotalTime=",mStatTotalTime);
    
    long slept=(mStatTotalTime==0)?0:
      ((mStatSleptTime*100) / mStatTotalTime);
   //Nog.msg('D',"   mStatSleptTime=",mStatSleptTime," slept=",slept,"%");
   
    long over=(mStatTotalTime==0)?0:
      ((mStatOverloadTime*100) / mStatTotalTime);
   //Nog.msg('D',"   mStatOverloadTime=",mStatOverloadTime," over=",over,"%");
   
   
   
    long overAvg=(mStatOverloadTimes==0)?0:
      (mStatOverloadTime/mStatOverloadTimes);
   //Nog.msg('D',"   mStatOverloadTimes=",mStatOverloadTimes," overAvg=",overAvg);
   
    
    //long workAvg=(mStatClockCnt==0)?0:
    //  (mStatWorkTime/mStatClockCnt);
    
    ////Nog.msg('D',"   mStatWorkTime=",mStatWorkTime);
    ////Nog.msg('D',"   mStatClockCnt=",mStatClockCnt," workAvg=",workAvg);
   
    
    
  //  long aWorkCnt=mStatClockCnt-mStatCanceledAddWord;
      
    //long aWorkAvg=(aWorkCnt==0)?0:
    //  (mStatAddWorkTime/(aWorkCnt));
    
    ////Nog.msg('D',"   mStatAddWorkTime=",mStatAddWorkTime);  
    ////Nog.msg('D',"   mStatCanceledAddWord=",mStatCanceledAddWord," aWorkCnt=",aWorkCnt," aWorkAvg=",aWorkAvg);
    
    long cyclesPerSec=(mStatTotalTime==0)?0:
    ((mStatClockCnt*10000)/mStatTotalTime);
   
    return 
    ""+mStatClockCnt+
    " s="+slept+"%"+
     "\no="+over+"%"+
    "oa="+overAvg+
    "\nspd="+cyclesPerSec+
    "\n"//+
   // "w="+workAvg//+
    //" a="+aWorkAvg+
    //"\n"+
    //"ca="+mStatCanceledAddWord+"";
     ;   
    
  }   
 
  public void logStat()
  {
   //Nog.msg('D',"GIThread stats: "+getStatString());     
  }
  
  /**
   * The thread loop function
   * this function will loop infinitly
   * calling at regular intervals the doTheJob function
   * an monitoring its delay in order to keep the game 
   * running always at the same speed
   */
  private void speedTest()
  {
    
    //Nog.msg('I',"GIThread.speedTest() started");           
    
    mBusyFlag=false;
    long timeAfter=System.currentTimeMillis();
      
    for (int i=0;i<1000;i++)
    { 
     try
      {
      	// compute time delay
      	long timeBefore=timeAfter;
        timeAfter=System.currentTimeMillis();
        long timeDelay=timeAfter-timeBefore;
        mStatTotalTime+=timeDelay;
       
        mStatClockCnt++;
        
      	 // request for additional work
        mHarness.requestPaint();
        Thread.currentThread().yield();
      } 
      catch(Exception e)
      {
       //Nog.msge('E',"GIThread.infiniteLoop() uncaught exception ",e);
        mHarness.setError("in speedTest. ex:\n"+e.getClass().getName()+"\n"+e.getMessage());
      }
    }
      
      
     //DEBUG for speed purpose:
     long cyclesPerSec=(mStatTotalTime==0)?0:
     ((mStatClockCnt*10000)/mStatTotalTime);
     //Nog.msg('I',"GIThread.speedTest() mStatTotalTime=",mStatTotalTime," mStatClockCnt=",mStatClockCnt," cyclesPerSec=",cyclesPerSec);           
 
     mHarness.setError("test\n:cps="+cyclesPerSec); 
     mHarness.requestPaint();
     Thread.currentThread().yield();
   
  }
        
  
  /**
   * The thread loop function
   * this function will loop infinitly
   * calling at regular intervals the doTheJob function
   * an monitoring its delay in order to keep the game 
   * running always at the same speed
   */
  private void infiniteLoop()
  {
    
   //Nog.msg('I',"GIThread.infiniteLoop() started");           
    
    mBusyFlag=false;
    long timeAfter=System.currentTimeMillis();
      
    while (!mHarness.mExitFlag)
    { 
      
      
      try
      {
      	
      
        
        //if (mClockCnt==20)
         // initStat();
          
        mHarness.updateKeyCodes();
        //DEBUG{
        logStat(); 
        //DEBUG}
        
       
        //if (mHarness.mInPaintWrap)
        //{  
        //    while (mHarness.mInPaintWrap)
        //   {
        //     //Nog.msg('D',"GIThread.waiting for paint to finish()...");
        //
        //      Thread.currentThread().sleep(mClockDelay/4);
        //    }
        ////Nog.msg('D',"GIThread paint seems to be finished now");
        // }
        
        
       //Nog.msg('D',"-----------------------------------------");
       //Nog.msg('D',"GIThread.calling step(",mClockCnt,")...");
      
        // now call the game manager
        //mBusyFlag=true;
        
        if (!mHarness.mErrorFlag) 
        {
          synchronized(mSemaphore)
		  {
        	mHarness.step(mClockCnt);
		  }
        }
        else
        {
         //Nog.msg('E',"now in error mode (msg=",mHarness.mErrorMsg,")");
        }
        //mBusyFlag=false;
        
        
        
        
        // request for additional work
        mHarness.requestPaint();
        Thread.currentThread().yield();
        
        
        // time managmenet
        
        long timeBefore=timeAfter;
        timeAfter=System.currentTimeMillis();
        
        // compute time delay
        long timeDelay=timeAfter-timeBefore;
       // mStatWorkTime+=timeDelay;
        
        mStatTotalTime+=timeDelay;
        
        //DEBUG{
        if (noDelay) // used to check game speed
        {
          mStatWorkTime+=timeDelay;
              
        }
        else 
        //DEBUG}	
        if (timeDelay<mClockDelay) // sleep to adjust game speed if required
        {
         // mHarness.log(" sleeping "+(clockDelay-timeDelay));
          Thread.currentThread().sleep(mClockDelay-timeDelay);
        
         //Nog.msg('T',"GIThread.  -sleeping:",mClockDelay-timeDelay);   
          mStatSleptTime+=mClockDelay-timeDelay;
          
          mStatWorkTime+=mClockDelay;
          
        }
        //DEBUG{
        else
        {
         //Nog.msg('T',"GIThread.  -overload:",-mClockDelay+timeDelay);   
          mStatOverloadTimes++;
          mStatOverloadTime=-mClockDelay+timeDelay;
          mStatWorkTime+=timeDelay;
      
        }
        //DEBUG}        
        mClockCnt++;
        //DEBUG{
        mStatClockCnt++;
        //DEBUG}
        
      } 
       
      catch(Exception e)
      {
        //Nog.msge('E',"GIThread.infiniteLoop() uncaught exception ",e);
     
        mHarness.setError("in infiniteLoop. ex:\n"+e.getClass().getName()+"\n"+e.getMessage());
    
      }
    } // while
   //Nog.msg('E',"GIThread. infiniteLoop() done"); 
    
    mHarness.done();
     
  }  
  
 /**
  * called by system when game starts
  */
  public void run()
   {
   //Nog.msg('I',"GIThread.run() started");
   infiniteLoop();
   //speedTest();
   //Nog.msg('I',"GIThread.run() done");
  }
  
  
  

  
  
  
  
}