// GITiledScreen.java
//---------------

package cw.GI;


import cw.log.Log;
/**
 *
 * This class is aimed to provide a tiled virtual screen.
 *
 * Its main advantage is to handle transparently a cache i.e. a buffer
 * to avoid redrawing the tiles individually. 
 * 
 * The buffer may be bigger or equal to the viewable area.
 * The move method() allows to transparently move into the tiled screen.
 * when the boundaries of the buffer are reached, the buffer is redrawn,
 * with the new position beeing at the center of the buffer
 * 
 * 
 */
public class GITiledScreen
{
    /** reference to the harness */
    GIHarness mHarness;
    
    // screen size in tiles
    public int mScreenSizeX;
    public int mScreenSizeY;
    
    
    // screen size in pixels
    public int mScreenSizeXxmTileSizeX;
    public int mScreenSizeYxmTileSizeY;
    
    
	
    
    // buffer size in tiles
    public int mBufferSizeX;
    public int mBufferSizeY;
    
    
    // displayable size 
    // in tiles
    public int mViewSizeX;
    public int mViewSizeY;
    
    
    // Tile size in pixels
    public int mTileSizeX;
    public int mTileSizeY;
    
    // tiles array
    public int mTiles[];

    // current view position 
    // in tiles
    public int mPosX;
    public int mPosY;
    
    // current view position 
    // in pixels
    public int mPosXxmTileSizeX;
    public int mPosYxmTileSizeY;
    
    
    // position of the buffer in the tiled screen 
    // in tiles
    public int mBufferPosX=0;
    public int mBufferPosY=0;
    
    
    // default tile image value
    public int mClearTile=1;

    
    /** used to indicate when the tiles array has changed
     * but the buffer has not been redrawn */
    boolean mDirtyCache;
    
    /**
     * 
     * @param pSizeX        tiled screen size (in tiles)
     * @param pSizeY
     * @param pBufferSizeX  buffer size (in tiles)
     * @param pBufferSizeY
     * @param pTileX        tile size (in pixels) 
     * @param pTileY
     * @param pViewSizeX.   real screen size (in pixels)
     * @param pViewSizeY
     */
    public void setSize(int pSizeX,int pSizeY,int pBufferDeltaX,int pBufferDeltaY,int pTileX,int pTileY,int pViewSizeX,int pViewSizeY)
    {
        
      //Nog.msg('I',"GITiledScreen.setSize() px=",pSizeX," py=",pSizeY, " tilex=",pTileX," tiley=",pTileY);
      mScreenSizeX=pSizeX;
      mScreenSizeY=pSizeY;
      
      
      
      
      mTileSizeX=pTileX;
      mTileSizeY=pTileY;    
      
      
      
      mScreenSizeXxmTileSizeX=mScreenSizeX*mTileSizeX;
      mScreenSizeYxmTileSizeY=mScreenSizeY*mTileSizeY;
      
      mTiles=new int[mScreenSizeX*mScreenSizeY];
    
      for (int i=0;i<mScreenSizeX;i++)
        for (int j=0;j<mScreenSizeY;j++)
        {
            mTiles[i+j*mScreenSizeX]=mClearTile;
        }
      
      
      mViewSizeX=(pViewSizeX+mTileSizeX-1)/mTileSizeX;
      mViewSizeY=(pViewSizeY+mTileSizeY-1)/mTileSizeY;

      
      mBufferSizeX=mViewSizeX+pBufferDeltaX;
      mBufferSizeY=mViewSizeY+pBufferDeltaY;    
      
      if (mBufferSizeX>mScreenSizeX) mBufferSizeX=mScreenSizeX;
      if (mBufferSizeY>mScreenSizeY) mBufferSizeY=mScreenSizeY;
      
      
      //Nog.msg('D'," mBufferSizeX=",mBufferSizeX,"  mBufferSizeY=",mBufferSizeY);
      
      
      //Nog.msg('D'," mViewSizeX=",mViewSizeX,"  mViewSizeY=",mViewSizeY);
      
    }
    
    
    
    /**
     *
     * init with a cache.
     * pImageIndex is the cache to be used.
     */
    
    public int mImageIndex;
    public void init( GIHarness pHarness,int pImageIndex,int x,int y)
    {
        //Nog.msg('I',"GITiledScreen.init() with cache. pImageIndex=",pImageIndex);
        
        mHarness=pHarness;

        mImageIndex=pImageIndex;
        
        
        int xSize=mBufferSizeX*mTileSizeX;
        int ySize=mBufferSizeY*mTileSizeY;
        //Nog.msg('I'," allocating image for buffer xSize=",xSize," ySize=",ySize); 
        mHarness.createImage(mImageIndex,xSize,ySize);
        mDirtyCache=true;
    
        
        mPosX=0;
        mPosY=0;
        
        move(x/mTileSizeX,y/mTileSizeY);
    
        
    }
    
    
    /**
     * this paint counter is used to know when the buffer
     * has been redraw.
     */
    public int mPaintCnt=0;
    
    
    /**
     * wrapper for drawTilesInBuffer(),using pixel position
     */
    public void drawInBuffer(int x1,int y1,int xd,int yd)
    {
    	//Nog.msg('D',"drawInBuffer() x1=",x1," y1=",y1," xd=",xd," yd=",yd);
    	
        int tileX=x1/mTileSizeX;
        int tileY=y1/mTileSizeY;
        int tileXe=(x1+xd-1)/mTileSizeX;
        int tileYe=(y1+yd-1)/mTileSizeY;
        drawTilesInBuffer(tileX,tileY,tileXe+1-tileX,tileYe+1-tileY);
    }
    
    /**
     * redraws a given area. the coordinates are in tiles.
     * 
     * this function has two internal flavors.
     * - non optimized (when mLoopUniverse. plus and it can handle transparent tiles)
     * - optimized 
     * 
     */
    public void drawTilesInBuffer(int x1,int y1,int xd,int yd)
    {
      
      //Nog.msg('D',"drawTilesInBuffer() x1=",x1," y1=",y1," xd=",xd," yd=",yd);
      int x2=x1+xd;
      int y2=y1+yd;
      
      
      
      // non optimized version
      if (mLoopUniverse) // && canHaveTransparentTiles
      {
      	// redraw the cache
      	for (int j=y1;j<y2;j++)
      	for (int i=x1;i<x2;i++)
      	{
      		int tile=getTile(i,j);
      		if (tile>=0)
      			mHarness.printImage((i-mBufferPosX)*mTileSizeX,(j-mBufferPosY)*mTileSizeY,tile,mImageIndex);
      	}
	  }
      else
      // optimized version.
      // also took off the test for transparent tiles
      {
      	int tileLineOffset=y1*mScreenSizeX+x1;
        int pixelY=(y1-mBufferPosY)*mTileSizeY;
      	int pixelXStart=(x1-mBufferPosX)*mTileSizeX;
      
      	for (int j=y1;j<y2;j++)
      	{
      		int tileOffset=tileLineOffset;
            int pixelX=pixelXStart;
            
      		for (int i=x1;i<x2;i++)
      		{
      			int tile=mTiles[tileOffset];
      			//DEBUG{
      			if (tile!=getTile(i,j)) throw new RuntimeException("tile offset computation failed. i="+i+" j="+j+" tileOffset="+tileOffset);
      			//DEBUG}
      			
      			mHarness.printImage(pixelX,pixelY,tile,mImageIndex);
      			
      			// DEBUG{
      			if ((i-mBufferPosX)*mTileSizeX!=pixelX) throw new RuntimeException("pixelX holds wrong value  pixelX="+pixelX+" not equal to j="+j+" x mTileSizeX"+mTileSizeX);
      			if ((j-mBufferPosY)*mTileSizeY!=pixelY) throw new RuntimeException("pixelY holds wrong value. pixelY="+pixelY+" not equal to j="+j+" x mTileSizeY"+mTileSizeY);
      			// DEBUG}
      			
      			tileOffset++;
      			pixelX+=mTileSizeX;
      		}
      		tileLineOffset+=mScreenSizeX;
      		pixelY+=mTileSizeY;
      	}	
      }
      
            
    }  

    /**
     * this function performs the actual paint
     */
    public void paint()
    {
    	
      // if the cache should be redraw, do it now	
      if (mDirtyCache) 
      {
        drawTilesInBuffer(mBufferPosX,mBufferPosY,mBufferSizeX,mBufferSizeY);
        mPaintCnt++;
        mDirtyCache=false;
      }
      
      //DEBUG{
      if (mPrintBufferAtPosX!=-(mPosX-mBufferPosX)*mTileSizeX) throw new RuntimeException("mPrintBufferAtPosX="+mPrintBufferAtPosX+" not equal to - (mPosX="+mPosX+" -mBufferPosX="+mBufferPosX+") x mTileSizeX="+mTileSizeX);
      if (mPrintBufferAtPosY!=-(mPosY-mBufferPosY)*mTileSizeY) throw new RuntimeException("mPrintBufferAtPosY="+mPrintBufferAtPosY+" not equal to - (mPosY="+mPosY+" -mBufferPosY="+mBufferPosY+") x mTileSizeY="+mTileSizeY);   
      //DEBUG}  
      
      
      // now print cache onto default output 
      mHarness.printImage(mPrintBufferAtPosX,mPrintBufferAtPosY,mImageIndex);
    }
 
    
    private int mPrintBufferAtPosX=0;
    private int mPrintBufferAtPosY=0;
    
    
    /**
     * this function moves the tiled screen view current position
     */
    public void move(int pDeltaX,int pDeltaY)
    {
        //Nog.msg('D',"GITiledScreen.move(",pDeltaX,",",pDeltaY,")");
        
        if (pDeltaX!=0)
        {
        	int newX= mPosX+pDeltaX;
        	if (newX>=0 && newX<=mScreenSizeX-mViewSizeX) mPosX+=pDeltaX;
        	mPosXxmTileSizeX=mPosX*mTileSizeX;
        }
        
        if (pDeltaY!=0)
        {
        	int newY= mPosY+pDeltaY;
        	if (newY>=0 && newY<=mScreenSizeY-mViewSizeY) mPosY+=pDeltaY;
        	mPosYxmTileSizeY=mPosY*mTileSizeY;
        }
        
        
        // should the buffer be redraw?
        
        //Nog.msg('D'," mPosX:",mPosX, " mBufferPosX:",mBufferPosX);
        //Nog.msg('D'," mPosX+mViewSizeX:",mPosX+mViewSizeX, " mBufferPosX+mBufferSizeX:",mBufferPosX+mBufferSizeX);
        //Nog.msg('D'," mPosY:",mPosY, " mBufferPosY:",mBufferPosY);
        //Nog.msg('D'," mPosY+mViewSizeY:",mPosY+mViewSizeY, " mBufferPosY+mBufferSizeY:",mBufferPosY+mBufferSizeY);
        
        if ((mPosX<mBufferPosX || mPosX+mViewSizeX>mBufferPosX+mBufferSizeX)
        		|| (mPosY<mBufferPosY || mPosY+mViewSizeY>mBufferPosY+mBufferSizeY))
        {
        	
        	
        	// the max position of the buffer pos
        	int mBufferMaxPosX=mScreenSizeX-mBufferSizeX;
        	int mBufferMaxPosY=mScreenSizeY-mBufferSizeY;
        	
        	// the deltaBuffer to add for a given position 
        	// to have the buffer in the middle
        	
        	
        	int mBufferMarginX=(mBufferSizeX-mViewSizeX)/2;
        	//Nog.msg('D',"mBufferMaxPosX:",mBufferMaxPosX," mBufferMarginX:",mBufferMarginX);
        	int mBufferMarginY=(mBufferSizeY-mViewSizeY)/2;
        	//Nog.msg('D',"mBufferMaxPosY:",mBufferMaxPosY," mBufferMarginY:",mBufferMarginY);
        	
        	
        	int newBufferX=mPosX-mBufferMarginX;
        	//Nog.msg('D',"newX before boundary check:",newBufferX);
        	int newBufferY=mPosY-mBufferMarginY;
        	//Nog.msg('D',"newY before boundary check:",newBufferY);
        	
        	
        	// compute new desired buffer position
        	newBufferX=(newBufferX<0)?0:((newBufferX>mBufferMaxPosX)?mBufferMaxPosX:newBufferX);
        	//Nog.msg('D',"newX:",newBufferX);
        	newBufferY=(newBufferY<0)?0:((newBufferY>mBufferMaxPosY)?mBufferMaxPosY:newBufferY);
        	//Nog.msg('D',"newY:",newBufferY);
        	
        	mBufferPosX=newBufferX;
        	mBufferPosY=newBufferY;
        	
        	drawTilesInBuffer(mBufferPosX,mBufferPosY,mBufferSizeX,mBufferSizeY);
        	//try
			//{
        	//Thread.sleep(500);
			//} catch(Exception e)
			//{ // ok
			//}
        	
        }
        
        // compute convenience data.
        
        // used by drawTiles 
        mPrintBufferAtPosX=-(mPosX-mBufferPosX)*mTileSizeX;
        mPrintBufferAtPosY=-(mPosY-mBufferPosY)*mTileSizeY;
       
        // used by the print image on buffer method
        mBufferPosXxmTileSizeX=mBufferPosX*mTileSizeX;
		mBufferPosYxmTileSizeY=mBufferPosY*mTileSizeY;
        
    }
    
    
    
    /** this parameter indicates if the universe should 'cycle' or not      */
     
    boolean mLoopUniverse=false;
     
    /**
     * return the tile located at a given position.
     * 
     */
     
    public int getTile(int x,int y)
    {
        int retVal;
        if (mLoopUniverse)
        {
        
            if (x>=mScreenSizeX) x=x%mScreenSizeX;
            else if (x<0) x=x%mScreenSizeX;
    
            if (y>=mScreenSizeY) y=y%mScreenSizeY;
            else if (y<0) y=y%mScreenSizeY;
            retVal=mTiles[x+y*mScreenSizeX];
            
            ////Nog.msg('D',"GITiledScreen.getTile(",x,",",y,") [",x+y*mSizeX,"] returns ",retVal);
    
        }
        else
        {
            if ((x>=mScreenSizeX) || (x<0) || (y>=mScreenSizeY) || (y<0) ) 
            {
                retVal=mClearTile;
                ////Nog.msg('D',"GITiledScreen.getTile(",x,",",y,") (out of universe) returns   ",retVal);
    
            }
            else 
            {
                retVal=mTiles[x+y*mScreenSizeX];
                ////Nog.msg('D',"GITiledScreen.getTile(",x,",",y,") [",x+y*mSizeX,"] returns ",retVal);
    
            }
            
                
        }
    
        return retVal;
        
    }
    
    /**
     * set the tiles internal array
     */
    public void setTiles(int pTiles[])
    {
        int read=0;
        
        for (int j=0;j<mScreenSizeY;j++)
            for (int i=0;i<mScreenSizeX;i++)
        
            {
            //Nog.msg('D',"GITiledScreen.setTiles() at (x=",i,",y=",j,") [",i+j*mSizeX,"] = ",pTiles[read]);
    
                mTiles[i+j*mScreenSizeX]=pTiles[read];
                read++;
                if (read>=mScreenSizeX*mScreenSizeY || read>=pTiles.length) return;
        
            }
        mDirtyCache=true;
    }
    
    /**
     * sets only a given tile.
     * x and y in tiles coordinates
     * image is the image index
     */
    public void setTile(int x,int y,int image)
    {
      mTiles[x+y*mScreenSizeX]=image;
      if(mDirtyCache==false)
      {
         mHarness.printImage(x*mTileSizeX,y*mTileSizeY,image,mImageIndex);
      }
      
    }
    
    /** holds the buffer position in world pixels */
    public int mBufferPosXxmTileSizeX;
    public int mBufferPosYxmTileSizeY;
    
	
    
    /**
     * prints a given image into the buffer
     * coordinates in world pixels
     */
   public void printImageInBuffer(int x,int y,int image)
   {
   	  mHarness.printImage(x-mBufferPosXxmTileSizeX,y-mBufferPosYxmTileSizeY,image,mImageIndex);
    
   }
   
}
