package cw.platf;
import cw.GI.*;
import cw.log.Log;
import java.util.Vector;

	

/**
 *
 * represent a game element to be generated 
 *
 */
public class GameData
{
  public int mType;
  public int mArgs[];
  public int mActivity;
  public GameData mLinks[];
  
  int mX;
  int mY;
  
  
  public final static int ACT_START=0;
  public final static int ACT_ALIVE=1;
  public final static int ACT_OUT=2;
  public final static int ACT_DIED=3;
  public int mPersistentData[];
  
  
  
  
  GameData(int type,int x,int y,int args[])
  {
    mType=type;
    mArgs=args;
    mActivity=ACT_START;
    mPersistentData=null;
    mLinks=null;
    mX=x;
    mY=y;
  }
  
  GameData(int type,int x,int y,int args[],GameData[] links)
  {
    this(type,x,y,args);
    mLinks=links;
  }
  
  
  public final static int PLAYER=0;
  public final static int BONUS_AXES=1;
  public final static int BONUS_LIFE=2;
  public final static int BAD_B=10;
  public final static int BAD_A=11;
  public final static int SNAKE=12;
  
  public final static int MENU_START=40;
  public final static int HMOVER=41;
  public final static int VUPMOVER=42;
  public final static int SWITCH=43;
  
  
  public final static int LINK=50;
  
  
  
  
  
  
  public void generate(Platf pGM)
  {
     mActivity=ACT_ALIVE;
     switch (mType)
     {
      
       case PLAYER:
         generatePlayer(pGM,mX,mY);
         break;
         
       case BONUS_AXES:
        generateBonusAxes(pGM,mX,mY);
         break;
        
       case BONUS_LIFE:
        generateBonusLife(pGM,mX,mY);
         break; 
         
       case BAD_B:
        generateBadB(pGM,mX,mY,mArgs[0],mArgs[1],mArgs[2]);
         break; 
       
       case BAD_A:
        generateBadA(pGM,mX,mY,mArgs[0]);
         break; 
         
       case MENU_START:  
         generateMenuStart(pGM,mX,mY);
         break;
       case HMOVER:  
         generateHorizontalMover(pGM,mX,mY,mArgs[0],mArgs[1],mArgs[2],mArgs[3]);
         break;
       case VUPMOVER:  
         generateVerticalUpMover(pGM,mX,mY,mArgs[0],mArgs[1],mArgs[2]);
         break;
       case SWITCH:  
         generateSwitch(pGM,mX,mY,mArgs[0]);
         break;   
      case LINK:  
      	 mActivity=ACT_START;
		generateLink(pGM,mLinks[0]);
		break;   

	 case SNAKE:  
      	
        generateSnake(pGM,mX,mY,mArgs[0],mArgs[1],mArgs[2]);
  
		break;   
	
         
         
       default: throw new RuntimeException("generate() found unknown type "+mType+" at position "+mX+","+mY);
     }
     
     
  }  

  //--+ +--------------------------------------------------------

  public void generatePlayer(Platf pGM,int pX,int pY)
  {
    
    //if (pGM.getFreeIndex()!=0) throw new RuntimeException("in generatePlayer(). player must be the first element inserted in the world");
    pGM.addElement(0,pGM.mPlayer=new GamePlayer(pGM));
    pGM.mPlayer.mX=pX;
    pGM.mPlayer.mY=pY;
  }
  
  
  public void generateBonusAxes(Platf pGM,int pX,int pY)
  {
      
      GameElement elem=new GameElement(pGM,"bonus",pX,pY,pGM.I_AXE_01);
      elem.setScript(new Script(elem,new int[] {
       
        Script.SET_IMG,pGM.I_AXE_00,
        Script.DELAY,4,
        Script.SET_IMG,pGM.I_AXE_03,
        Script.DELAY,4,
        Script.SET_IMG,pGM.I_AXE_02,
        Script.DELAY,4,
        Script.SET_IMG,pGM.I_AXE_01,
        Script.DELAY,4,
        Script.GOTO,0,
        Script.END}));   
        
      elem.setBonus(0,5,0);
      elem.mIsKillable=false;
      elem.mDirectDraw=true;
      pGM.addElement(elem);
      
      elem.mGameData=this;
      

   }
   
   
   public void generateBonusLife(Platf pGM,int pX,int pY)
   {
      GameElement elem=new GameElement(pGM,"bonus",pX,pY,pGM.I_BONUS_00);
      elem.setScript(new Script(elem,new int[] {
       
        Script.SET_IMG,pGM.I_BONUS_00,
        Script.DELAY,4,
        Script.SET_IMG,pGM.I_BONUS_01,
        Script.DELAY,4,
        Script.GOTO,0,
        Script.END}));   
        
      elem.setBonus(5,0,0);
  
      elem.mIsKillable=false;
      
      elem.mDirectDraw=true;
      pGM.addElement(elem);
      
      elem.mGameData=this;
      
      
    } 
    
    
   public void generateBadB(Platf pGM,int pX,int pY,int lX1,int lX2,int speed)
   {
   
    
      int delayToX2=((lX2-pX)*2)/speed;
      int delayBetweenX1AndX2=((lX2-lX1)*2)/speed;
      
    
      GameElement elem=new GameElement(pGM,"badb",pX,pY,pGM.I_BAD_B_R_00);
      elem.initAutoAnims(3);
      elem.setAutoAnim(0,
        new int[] {
        2,pGM.I_BAD_B_R_00,
        2,pGM.I_BAD_B_R_01,
        -1});
     
      elem.setAutoAnim(1,
        new int[] {
        2,pGM.I_BAD_B_L_00,
        2,pGM.I_BAD_B_L_01,
        -1});
        
      
      
      elem.setScript(new Script(elem,new int[] {
        Script.SET_AANIM,0,
        Script.SET_AMOVE,2,speed,0,0,
        Script.DELAY,delayToX2,
        
        Script.SET_AANIM,1,
        Script.SET_AMOVE,2,-speed,0,0,  
        Script.DELAY,delayBetweenX1AndX2,
        
        Script.SET_AANIM,0,
        Script.SET_AMOVE,2,speed,0,0,  
        Script.DELAY,delayBetweenX1AndX2,
        
        
        Script.GOTO,9,
        Script.END}));   
        
        
       //elem.mIsTouchable=false; 
      pGM.addElement(elem);
      elem.mGameData=this;
  }
   
   
   
   
   
   
   
   public void generateSnake(Platf pGM,int pX,int pY,int lX1,int lX2,int speed)
   {
   
    
      int delayToX2=((lX2-pX)*2)/speed;
      int delayBetweenX1AndX2=((lX2-lX1)*2)/speed;
      
    
      GameElement elem=new GameElement(pGM,"badb",pX,pY,pGM.I_BAD_C_R_00);
      elem.initAutoAnims(3);
      elem.setAutoAnim(0,
        new int[] {
        2,pGM.I_BAD_C_R_00,
        2,pGM.I_BAD_C_R_01,
        -1});
     
      elem.setAutoAnim(1,
        new int[] {
        2,pGM.I_BAD_C_L_00,
        2,pGM.I_BAD_C_L_01,
        -1});
        
      
      
      elem.setScript(new Script(elem,new int[] {
        Script.SET_AANIM,0,
        Script.SET_AMOVE,2,speed,0,0,
        Script.DELAY,delayToX2,
        
        Script.SET_AANIM,1,
        Script.SET_AMOVE,2,-speed,0,0,  
        Script.DELAY,delayBetweenX1AndX2,
        
        Script.SET_AANIM,0,
        Script.SET_AMOVE,2,speed,0,0,  
        Script.DELAY,delayBetweenX1AndX2,
        
        
        Script.GOTO,9,
        Script.END}));   
        
        
       //elem.mIsTouchable=false; 
      pGM.addElement(elem);
      elem.mGameData=this;
  }
    
   /**
    * generates a hiding monster. 
    */
    public void generateBadA(Platf pGM,int pX,int pY,int delay)
   {
      
      GameElement elem=new GameElement(pGM,"bad",pX,pY,pGM.I_BAD_A_02);
      elem.setScript(new Script(elem,new int[] {
        Script.SET_TOUCHABLE,0,
        Script.SET_IMG,pGM.I_BAD_A_02,
        Script.DELAY,delay,
        
        Script.SET_TOUCHABLE,1,
        Script.SET_IMG,pGM.I_BAD_A_01,
        Script.DELAY,6,
        Script.SET_IMG,pGM.I_BAD_A_00,
        Script.DELAY,6,
        Script.SET_IMG,pGM.I_BAD_A_01,
        Script.DELAY,6,
        Script.SET_IMG,pGM.I_BAD_A_00,
        Script.DELAY,6,
        
        
        Script.GOTO,0,
        Script.END}));   
        
      elem.mDirectDraw=true;
      pGM.addElement(elem);
      elem.mGameData=this;
      
    } 
    /*
     * generates a menu for allowing the user to exit the game
     * or to start playing.
     * 
     * only to be used in the intro world.
     */
    public void generateMenuStart(Platf pGM,int pX,int pY)
   {
      
     {
      GameElement elem=new GameElement(pGM,"playquit",15,20,pGM.I_PLAY_QUIT);
      pGM.addElement(elem);
    }
    
   
    {
      GameElement elem=new GameElementMenu(pGM,"menu",5,20,2,10)
      {
        public void menuSelected(int pPos)
        {
          if (pPos==0)
          {
            mGM.mWorldMgr.loadWorld2();
            mGM.mWorldMgr.startWorld();    
          }
          else
          {
            mH.stop();
          }

        }
      };
      pGM.addElement(elem);
    }
      
   } 
    
     public void generateHorizontalMover(Platf pGM,int pX,int pY,int lP1,int lP2,int speed,int image)
   {
      int delay=3;
   
      int delayToP2=((lP2-pX)*delay)/speed;
      int delayBetweenP1AndP2=((lP2-lP1)*delay)/speed;
    
      GameElement elem=new GameElementMover(pGM,"bada",pX,pY,image);
     
     
      pGM.addElement(elem);
     
      elem.setScript(new Script(elem,new int[] {
        Script.SET_AMOVE,delay,speed,0,0,
        Script.DELAY,delayToP2,
        
        Script.SET_AMOVE,delay,-speed,0,0,  
        Script.DELAY,delayBetweenP1AndP2,
        
        Script.SET_AMOVE,delay,speed,0,0,  
        Script.DELAY,delayBetweenP1AndP2,
        
        Script.GOTO,7,
        Script.END}));   
        
   
	
      
      
      elem.mGameData=this;
      elem.mIsKillable=false;
  }
     
     
   public void generateVerticalUpMover(Platf pGM,int pX,int pY,int lP,int speed,int image)
   {
      int delay=1;
   
      int delayToP=((lP-pY)*delay)/speed;
      if (delayToP<0) delayToP=-delayToP;
    
      GameElement elem=new GameElementMover(pGM,"bada",pX,pY,image);
     
      elem.setInScreen(pX,pY,pX,pY+lP*speed);
      pGM.addElement(elem);
     
      elem.setScript(new Script(elem,new int[] {
        Script.SET_AMOVE,0,0,delay,speed,
        Script.DELAY,delayToP,
        
        Script.GOTO_XY,pX,pY,  
        Script.GOTO,0,
        Script.END}));   
        
   
	
      
      
      elem.mGameData=this;
      elem.mIsKillable=false;
  }
   
   
   public void generateSwitch(Platf pGM,int pX,int pY,int way)
   {
     
    
      GameElement elem=new GameElementMover(pGM,"switch",pX,pY,pGM.I_SWITCH_ON,way)
	  {
	    public void onFire()
	    {
	      if (mImage==mGM.I_SWITCH_OFF)
	      {
	         mImage=mGM.I_SWITCH_ON;
	      }
	      else
	      {
	      	 mImage=mGM.I_SWITCH_OFF;
	      }
	    }
	  };
     
     
      pGM.addElement(elem);
      
      elem.mGameData=this;
      elem.mIsKillable=false;
  } 
   
   
   
    public void generateLink(Platf pGM,GameData linkedObject)
    {
      
		// this tests are copied from GameWorldMgr
		// they should be inside the called generate() function
		// but for speed opt. they are moved into the caller
        if (linkedObject.mActivity==GameData.ACT_START || linkedObject.mActivity==GameData.ACT_OUT)
        {
        	linkedObject.generate(pGM);
        }
    }
}

