// GameElement.java
//-----------

package cw.platf;
import cw.GI.*;
import cw.log.Log;

/*
 *
 * Parameters
 * Constructors
 * Flow
 * DirectDraw
 * ImageAnim 
 * AutoMove
 * AutoAnim
 */
 class GameElement
{

  //--+ Data +-------------------------------------------------------
  
  
  
  
  //--+ parameters +---------------
  
  String mName; // used for debug only
  
   // something will happen when the player touch this element
  public boolean mIsTouchable=true; 
  boolean mIsKillable; // can be kill when an axe touch it
  
  // bonus parameters use setBonus()
  public int mBonusLife=0;
  public int mBonusAxes=0;
  public int mMalus=-4;
  
  public boolean mCheckForInWindow=true;
  
  // Image to be displayed and size. use setImage() to set thoses.
  int mImage; 
  int mSizeX;
  int mSizeY;
  
  // position 
  int mX; 
  int mY;
  

  
  
  //--+ connection +------------------ 
  Platf mGM; // ref to the game manager
  GIHarness mH; // shortcut ref to the  harness
  Script mScript=null; // a script interpreter
  GameData mGameData;
  

  //--+ variables +--------------------  
  boolean mDeadFlag; // when this flag is set, the GameManager dispose this element 
  int mDeltaX=0;
  int mDeltaY=0;
  
  
  
  
  /**
   * set the image and update size according to image
   *
   */
  void setImage(int pImage)
  {
  	
  	//DEBUG{
  	if (mGM.mH.mImages[pImage]==null) throw new RuntimeException("empty image "+pImage);
  	//DEBUG}
    mSizeX=mGM.mH.mImages[pImage].mX;
    mSizeY=mGM.mH.mImages[pImage].mY;
    mImage=pImage;
  }  
  
  
  
  
  //--+ construction +------------------------------------------
  GameElement(Platf pGM,String pName,int pX,int pY)
  {
    mGameData=null;
    mName=pName;
    mGM=pGM;
    mH=pGM.mH;
    mX=pX;
    mY=pY;
    
   
    mImage=0;
    mSizeX=0;
    mSizeY=0;
    
    
    mDeadFlag=false;
    mImageAnimFlag=false;
    mIsKillable=true;
    
    
    setBonus(0,0,5);
    
    
    // set the 'in screen' area
    setInScreen(mX,mY,mX+pX,mY+pY);
    
    
  }
  
  

  GameElement(Platf pGM,String pName,int pX,int pY,int pImage)
  {
    this(pGM,pName,pX,pY);
    setImage(pImage);
    
  }
 
 
  
  
  
 //--+ flow +----------------------------------------------------------------
   
  
  public void manage()
  {
   
    if (mScript!=null) mScript.process();
    
    if (mAutoAnim)  processAutoAnim();
    if (mAutoMove) processAutoMove();
    
    mX+=mDeltaX;
    mY+=mDeltaY;
    
    
    if (mIsTouchable) manageTouchable();
    
    if (mCheckForInWindow) 
    {
      if (!isInScreen())
      {
        out();
        
        
      }
    }
     
    
    
  }
  
  
  
  	
  public void manageTouchable()
  {
  
	if(mGM.mPlayer!=null)
    {
      if (isInRect(mGM.mPlayer.mX,mGM.mPlayer.mY,5,10,0))
      {
        touched();
      }
    }
  }
      
  public void touched()
  {
    //Nog.msg('T',"touched()");
    
        if (mMalus>0)
        {
          mGM.mPlayer.looseEnergy(mMalus);
         
        }
        else if (mBonusLife>0)
     
        {
          mGM.mPlayer.earnEnergy(mBonusLife);
          die();
        }
        else if (mBonusAxes>0)
     
        {
          mGM.mPlayer.earnAxes(mBonusAxes);
          die();
        }
        
  }
  
  
  
   
   
  public void paint()
  //-------------------
  // called by gamemanager to display the element
  {
    if (mImageAnimFlag) processImageAnim();
    
     
    if (mDirectDraw)
    {  
      manageDirectDraw(); 
    }
    else
    {
      //DEBUG{
      try
	  {
      //DEBUG}	
      	mH.printImage(mX-mGM.mScreenX,mY-mGM.mScreenY,mImage);
	  //DEBUG{
	  }
      catch (Exception e)
	  {
      	throw new RuntimeException("exception "+e.getClass().getName()+" msg="+e.getMessage()+" in game element "+getClass().getName()); 
	  }
      //DEBUG}
    }
  }
  
  //--+ direct draw stuff +----------------------------------------------------
  
  
  int mImageDirectDrawed=0;
  int mImageDirectDrawedX=0;
  int mImageDirectDrawedY=0;
  
  int mTiledScreenPaintCnt=0;
  boolean mDirectDraw=false;
  
  
  public void manageDirectDraw()
  //-------------------
  // called by gamemanager to display the element
  // with the direct draw into the tiled screen buffer
  {
    
      if (mTiledScreenPaintCnt!=mGM.mTiledScreen.mPaintCnt)
      {
        //Nog.msg('F',"manageDirectDraw() redraw due to tiledScreen repaint");
        directDrawPrint();  
      }
      else if (mImageDirectDrawed!=mImage)
      {
        //Nog.msg('F',"manageDirectDraw() redraw due to image change");
       
        directDrawClear();
        directDrawPrint();  
        
      }
      else if ( mImageDirectDrawedX!=mX || mImageDirectDrawedY!=mY)
      {
      	//Nog.msg('F',"manageDirectDraw() redraw due to image position change");
       
        directDrawClear();
        directDrawPrint();          
      }
  }
   
   
  /**
   *
   * redraw the background  at previous direct draw position
   * in order to erase the stuff previously written
   *
   */
  public void directDrawClear()
  {        
      mGM.mTiledScreen.drawInBuffer(mImageDirectDrawedX,mImageDirectDrawedY,mSizeX,mSizeY);
  }
  
  public void directDrawPrint()
  {
  	  mImageDirectDrawed=mImage;
      mTiledScreenPaintCnt=mGM.mTiledScreen.mPaintCnt;
      mGM.mTiledScreen.printImageInBuffer(mX,mY,mImage);
      //mH.printImage(mX,mY,mImage,mGM.mTiledScreen.mImageIndex);
      mImageDirectDrawedX=mX;
      mImageDirectDrawedY=mY;
  }
  
  
  
  //--+ image anim stuff +-----------------------------------------------
  
  int mImageAnimArray[];
  int mImageAnimPos;
  int mImageAnimCnt;
  boolean mImageAnimFlag=false;
  
  public void setImageAnim(int pImageAnimArray[])
  {
    //Nog.msg('D',"setImageAnim()");
    mImageAnimArray=pImageAnimArray;
    mImageAnimPos=0;
    mImageAnimCnt=0;
    mImageAnimFlag=true;
  }
  
   
  public void processImageAnim()
  {
   
  
     
    //Nog.msg('D',"processImageAnim() mImageAnimCnt=",mImageAnimCnt);

    if (--mImageAnimCnt<=0)
    {
      
    
      
      while (true)
      {
        
        int code=mImageAnimArray[mImageAnimPos];
        //Nog.msg('D',"mImageAnimPos=",mImageAnimPos," code=",code);
    
        
        if (code>0)
        {
          mImageAnimCnt=code;
          mImage=mImageAnimArray[mImageAnimPos+1];
          //Nog.msg('D'," set image to ",mImage);
          mImageAnimPos+=2; 
          break;
        }
        else if (code==-1) 
        {
          mImageAnimPos=0; // return to 
          //Nog.msg('D'," restart");
   
        }
        else if (code==-2) 
        {
        	mImageAnimFlag=false; 
          mImage=mImageAnimArray[mImageAnimPos+1];     
          //Nog.msg('D'," set image to ",mImage," and end auto anim");
          mImageAnimPos+=2;
          break;
        }
        else
        {
        	throw new RuntimeException("unknown code "+code+" in processAutoAnim for element "+mName+" mImageAnimPos="+mImageAnimPos);
        }
      }
    }
  }
  
  
 
 //--+ auto move stuff +----------------------------------------
 
 
 int mAutoCntX;
 int mAutoCntY;
 int mAutoStepX;
 int mAutoStepY;
 
 int mAutoCurrentCntX;
 int mAutoCurrentCntY;
 
 boolean mAutoMove=false;
 
 public void initAutoMove(int cx,int sx,int cy,int sy)
 {
  
   // this is a small fix to speed up
   if (sx==0) cx=999;
   if (sy==0) cy=999;
  
   mAutoCntX=cx;
   mAutoCntY=cy;
   mAutoStepX=sx;
   mAutoStepY=sy;
   
   mAutoCurrentCntX=0;
   mAutoCurrentCntY=0;
   
   mAutoMove=true;
  
   //Nog.msg('D',"autoMove: cntX=",mAutoCntX," stepX=",mAutoStepX," cntY=",mAutoCntY," stepY=",mAutoStepY);
 }
 
 
 public void processAutoMove()
 {
  
    
   if (++mAutoCurrentCntX>=mAutoCntX)
   {
     mDeltaX=mAutoStepX;
     mAutoCurrentCntX=0;
   }
   else
   {
     mDeltaX=0;
   } 
   
   if (++mAutoCurrentCntY>=mAutoCntY)
   {
     mDeltaY=mAutoStepY;
     mAutoCurrentCntY=0;
   }
   else
   {
     mDeltaY=0;
   } 
  
 }
 
 
  //--+ auto anim stuff +---------------------------------------
 
 int mPrevAnimMode;
  int mAnimMode;
  boolean mAutoAnim=false;
  int [][] mAutoAnims;
  
  
  
  
  public void initAutoAnims(int max)
  {
    mAutoAnims=new int[max][];
    mAutoAnim=true;
    mPrevAnimMode=-1;
    mAnimMode=0;
    
  }
  
  public void setAutoAnim(int index,int[] data)
  {
    
    mAutoAnims[index]=data;
  }
  
  public void processAutoAnim()
  {
  
     
    //Nog.msg('D',"processAutoAnim()");
   
    if (mPrevAnimMode!=mAnimMode)
    {
      
      setImageAnim(mAutoAnims[mAnimMode]);
      mPrevAnimMode=mAnimMode;
    }
  }
  
  
  
  //--+ dying stuff +----------------------------------------------------------
  
 
  /**
   * drop this element because it died
   *
   */
  public void die()
  {
    
    //Nog.msg('I',"die() for elem=",mName);
   
     
    if (mIsTouchable)
    {
       if(mMalus>0) createEjectAnim(this,(mGM.mPlayer!=null)?mGM.mPlayer.mWay:1);
       else if(mBonusLife>0 || mBonusAxes>0) createBonusAnim(this);
    }
      
    if (mGameData!=null)
    {
      mGameData.mActivity=mGameData.ACT_DIED;
    }
      
    requestDestroy();
  }
  
  /**
   * drop this element because it went out of the screen or
   * sg similar.
   *
   */
  public void out()
  {
    
    //Nog.msg('I',"out() for elem=",mName);
   
    if (mGameData!=null)
    {
       mGameData.mActivity=mGameData.ACT_OUT;
    }
    requestDestroy();
  }      
     
  
  /**
   * inform the game manager that this element should be
   * disposed
   */
  void requestDestroy() 
  {
    //Nog.msg('I',"requestDestroy(",mName,")");
    
    //ASSERT{
    if (mGameData!=null)
    {
      if (mGameData.mActivity==mGameData.ACT_ALIVE)
        throw new RuntimeException("requestDestroy() for element "+mName+" but GameData.mActivity was not updated");
    }
    //ASSERT}
    mDeadFlag=true;
  } 
  
  
  /**
   * destructor. called by game manager just before loosing
   * its reference to this game element
   */
  public void done()
  {
    if (mDirectDraw)
    {
      directDrawClear(); 
    }
    mH=null;
    mGM=null;
    if (mScript!=null)
    {
      mScript.done();
      mScript=null;
    }
    
    mGameData=null;
  }
  
  //--+ ready to use animations +---------------------------------------------
  
  
  
  
  
  public static GameElement createEjectAnim(GameElement e,int way)
  {
    //Nog.msg('D',"createEjectAnim(",e.mName,",",way,")");
    GameElement elem=new GameElement(e.mGM,"dyed",e.mX,e.mY,e.mImage);
    elem.mIsTouchable=false;
   
    elem.setScript(new Script(elem,new int[] {
        Script.SET_DXY,way,-4,
        Script.DELAY,1,
        Script.SET_DXY,way,-2,
        Script.DELAY,1,
        Script.SET_DXY,way,+2,
        Script.DELAY,1,
        Script.SET_DXY,way,+4,
        Script.DELAY,2,
        Script.SET_DXY,way,+10,
        Script.DELAY,10,
 
        Script.DIE,
        Script.END}));     

   
        
    e.mGM.addElement(elem);
    return elem;      
    
  }
  
  public static  GameElement createBonusAnim(GameElement e)
  {
    //Nog.msg('D',"createBonusAnim("+e.mName+")");
    GameElement elem=new GameElement(e.mGM,"dyed",e.mX,e.mY,e.mImage);
    elem.mIsTouchable=false;
    
    elem.setScript(new Script(elem,new int[] {
        Script.SET_DXY,1,-2,
        Script.DELAY,3,
        Script.SET_DXY,-1,-2,
        Script.DELAY,3,
        Script.SET_DXY,1,-2,
        Script.DELAY,3,
        Script.DIE,
        Script.END})); 
        
    e.mGM.addElement(elem);
    return elem;      
        
  }
  
  
  //--+ miscs +----------------------------------------------------------------
   /**
   * set the interpreted script to s
   *
   */
  public void setScript(Script s)
  {
    mScript=s;  
  }
  
  
  /**
   * returns true when the element is inside the screen
   *
   */
   public int mInScreenX1=0;
   public int mInScreenY1=0;
   public int mInScreenX2=0;
   public int mInScreenY2=0;
   
  
   public void setInScreen(int xa,int ya,int xb,int yb)
   {
   	if (xa<xb) {mInScreenX1=xa;mInScreenX2=xb;} else {mInScreenX1=xb;mInScreenX2=xa;}
   	if (ya<yb) {mInScreenY1=ya;mInScreenY2=yb;} else {mInScreenY1=yb;mInScreenY2=ya;}
   	//Nog.msg('D'," mInScreenX1=",mInScreenX1," mInScreenY1=",mInScreenY1," mInScreenX2=",mInScreenX2," mInScreenY2=",mInScreenY2);
   
   }
   
   public boolean isInScreen()
  {
     
   /* int rX1=mInScreenX1-mGM.mScreenX;
    int rY1=mInScreenY1-mGM.mScreenY;
    int rX2=mInScreenX2-mGM.mScreenX;
    int rY2=mInScreenY2-mGM.mScreenY;

   	
    //Nog.msg('D',"isInScreen() rX1=",rX1," rY1=",rY1," rX2=",rX2," rY2=",rY2);
*/
    //Nog.msg('D'," mInWindowX1=",mGM.mInWindowX1," mInWindowY1=",mGM.mInWindowY1," mInWindowY1=",mGM.mInWindowY1," mInWindowY2=",mGM.mInWindowY2);
    
    boolean retVal=
     (mInScreenX1<mGM.mInWindowX2) && (mInScreenX2>mGM.mInWindowX1) &&
     (mInScreenY1<mGM.mInWindowY2) && (mInScreenY2>mGM.mInWindowY1);
       
    //Nog.msg('D'," retVal=",retVal);
    return retVal;  
  }
  
  
  public void setBonus(int bl,int ba,int m)
  {
     mBonusLife=bl;
     mBonusAxes=ba;
     mMalus=m;
     //Nog.msg('D',"setBonus() bl=",mBonusLife," ba=",mBonusAxes," m=",mMalus);
    
  }
  
  //--+ utils +------------------------------------------------------------------------
   
   
   
   
  
  
  int getTileAt(int dx,int dy)
  { 
    int inWorldX=(mX+dx)/mGM.mTiledScreen.mTileSizeX;
    int inWorldY=(mY+dy)/mGM.mTiledScreen.mTileSizeY;
    int retVal=mGM.mTiledScreen.getTile(inWorldX,inWorldY);
    //Nog.msg('I',"getTileAt() inWorldX=",inWorldX," inWorldY=",inWorldY," returns ",retVal);
    return retVal;
  }
    
  boolean isThereGround(int dx,int dy)
  {  
    int tile=getTileAt(dx,dy);
    return tile>=Platf.GROUND_START && tile <=Platf.GROUND_END;
  }
  
  boolean isThereScale(int dx,int dy)
  {  
   int tile=getTileAt(dx,dy);
   return tile>=Platf.SCALE_START  && tile<=Platf.SCALE_END ; 
  }
  
  boolean isThereGroundOrTopScale(int dx,int dy)
  {  
    int tile=getTileAt(dx,dy);
    return  tile==Platf.I_SCALE_01 || (tile>=Platf.GROUND_START && tile <=Platf.GROUND_END); 
  }
  
  boolean isThereGroundOrScale(int dx,int dy)
  {  
    int tile=getTileAt(dx,dy);
    return (tile>=Platf.SCALE_START  &&  tile<=Platf.SCALE_END) || (tile>=Platf.GROUND_START && tile <=Platf.GROUND_END); 
  }
  
     
     
     
      /**
   *  return true when it overlaps with another element e.
   *  delta is for expanding the tolerance.
   *
   */
  public boolean overlap(GameElement e,int delta)
  {
    return isInRect(e.mX,e.mY,e.mSizeX,e.mSizeY,delta);
  }
  
  
  /**
   * returns true when the elemnt area is inside x,y,x+dx,y+dy.
   *
   */
  public boolean isInRect(int x,int y,int dx,int dy,int delta)
  {
    ////Nog.msg('D',"isInRect() x=",x," y=",y," dx=",dx," dy=",dy);
    ////Nog.msg('D',"mX=",mX," mY=",mY," mSizeX=",mSizeX," mSizeY=",mSizeY);
    
    
    boolean retVal= ((x+dx+delta>mX) && (x<mX+mSizeX+delta) &&
       (y+dy+delta>mY) && (y<mY+mSizeY+delta));
       
    ////Nog.msg('D',"delta=",delta," retVal=",retVal);
    return retVal;  
  }
  
  
  
  
  
}


 
